# Copyright 2021 Jacob Baumbach
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Coordinator between different collections of types."""
from typing import Any
from typing import List
from typing import Optional
from typing import Type

from adorn.exception.type_check_error import TypeCheckError
from adorn.unit.unit import Unit


class Orchestrator:
    """Coordinate actions between a collection of :class:`~adorn.unit.unit.Unit`"""

    units: List[Unit]
    """a collection of collection of types"""

    def contains(self, cls: Type) -> bool:  # pragma: no cover
        """Check if ``Type`` exists in the ``Orchestrator``

        Args:
            cls (Type): a type that is being checked to see if
                it exists in the ``Orchestrator``

        Raises:
            NotImplementedError: This method must be overriden by a subclass
        """
        raise NotImplementedError

    def get(self, cls: Type) -> Unit:  # pragma: no cover
        """Finds the :class:`~adorn.unit.unit.Unit` associated with `cls`

        Args:
            cls (Type): a type that you want the associated ``Unit`` of

        Raises:
            NotImplementedError: This method must be overriden by a subclass
        """
        raise NotImplementedError

    def type_check(
        self, cls: Type, obj: Any
    ) -> Optional[TypeCheckError]:  # pragma: no cover
        """Check if `obj` can be converted to type of `cls`

        Args:
            cls (Type): the type ``obj`` would be converted into
            obj (Any): an instance to be converted into type ``cls``

        Raises:
            NotImplementedError: This method must be overriden by a subclass
        """
        raise NotImplementedError

    def from_obj(self, cls: Type, obj: Any) -> Any:  # pragma: no cover
        """Generate an instance of type `cls` from `obj`

        Args:
            cls (Type): the type ``obj`` would be converted into
            obj (Any): an instance to be converted into type ``cls``

        Raises:
            NotImplementedError: This method must be overriden by a subclass
        """
        raise NotImplementedError



# Copyright 2021 Jacob Baumbach
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Handle an argument for a parameter specified in the constructor of a class"""
from collections import defaultdict
from sys import version_info
from typing import Any
from typing import Dict
from typing import Generic
from typing import Optional
from typing import Type
from typing import TYPE_CHECKING
from typing import TypeVar
from typing import Union

if (version_info.major > 3) or (version_info.major == 3 and version_info.minor >= 8):
    from typing import Literal
else:
    from typing_extensions import Literal  # noqa: F401 # pragma: no cover

if TYPE_CHECKING:  # pragma: no cover
    from adorn.orchestrator.orchestrator import Orchestrator

from adorn.data.parameter import Parameter
from adorn.data.constructor import Constructor
from adorn.exception.type_check_error import (
    ExtraLiteralError,
    KeyValueError,
    MalformedDependencyError,
    MalformedLiteralError,
    MissingDependencyError,
    MissingLiteralError,
    TooDeepLiteralError,
    TypeCheckError,
    UnaryLiteralError,
)
from adorn.params import Params
from adorn.unit.complex import _T, Complex
from adorn.unit.simple import Simple

DictStrStr = TypeVar("DictStrStr", bound=Literal[dict()])


class ParameterValue(Simple):
    """:class:`~adorn.unit.unit.Unit` for handling an argument for a parameter specified in a constructor

    :class:`~adorn.data.parameter.Parameter` contains the type and state from the constructor,
    which ``ParameterValue`` allows you to utilize when type checking or instantiating the
    argument associated with the parameter
    """  # noqa: B950

    _registry = defaultdict(dict)

    def contains(self, obj: Any, orchestrator: "Orchestrator") -> bool:
        """Check if the ``obj`` is an instance of :class:`~adorn.data.parameter.Parameter`

        Args:
            obj (Any): potentially an instance of
                :class:`~adorn.data.parameter.Parameter`
            orchestrator (Orchestrator): container of all types, typically
                used to recurse nested types

        Returns:
            bool: if ``True``, ``obj`` is an instance of
                :class:`~adorn.data.parameter.Parameter`
        """  # noqa: B950
        return isinstance(obj, Parameter) and (self.get(obj, orchestrator) is not None)

    def get(self, key: Any, orchestrator: "Orchestrator") -> Optional["ParameterValue"]:
        """Get the relevant class to handle the given :class:`~adorn.data.parameter.Parameter` instance.

        Args:
            key (Any): potentially an instance of
                :class:`~adorn.data.parameter.Parameter``
            orchestrator (Orchestrator): container of all types, typically
                used to recurse nested types

        Returns:
            Optional[ParameterValue]: if ``key`` is an instance of
                :class:`~adorn.data.parameter.Parameter` then a subclass of
                ``ParameterValue`` will be returned, otherwise ``None``
        """  # noqa: B950
        if not isinstance(key, Parameter):
            return None
        # we default all Parameter's to use Identity
        return self._get(key=key, orchestrator=orchestrator) or (
            self._registry[ParameterValue]["identity"]
            if orchestrator.contains(key.cls)
            else None
        )


@ParameterValue.register("identity")
class Identity(ParameterValue):
    """Ignore the additional state in the :class:`~adorn.data.parameter.Parameter`

    We extract the type information from :class:`~adorn.data.parameter.Parameter`
    and pass off the object back to the :class:`~adorn.orchestrator.Orchestrator`
    """

    @classmethod
    def _from_obj(
        cls, target_cls: Parameter, orchestrator: "Orchestrator", obj: Dict[Any, Any]
    ) -> Any:
        """Delegate to the :class:`~adorn.orchestrator.Orchestrator` to instantiate the argument, ``obj``

        Args:
            target_cls (Parameter): instance of
                :class:`~adorn.data.parameter.Parameter`
            orchestrator (Orchestrator): container of all types, used to instantiate
                the argument
            obj (Any): an instance, containing the information to instantiate
                an instance of :attr:`~adorn.data.parameter.Parameter.cls`

        Returns:
            Any: An instance of :attr:`~adorn.data.parameter.Parameter.cls`
        """  # noqa: B950, RST304
        return orchestrator.from_obj(target_cls.cls, obj)

    @classmethod
    def _type_check(
        cls, target_cls: Parameter, orchestrator: "Orchestrator", obj: Dict[Any, Any]
    ) -> Optional[TypeCheckError]:
        """Delegate to the :class:`~adorn.orchestrator.Orchestrator` to type check the argument, ``obj``

        Args:
            target_cls (Parameter): instance of
                :class:`~adorn.data.parameter.Parameter`
            orchestrator (Orchestrator): container of all types, used to type
                check the argument
            obj (Any): an instance, containing the information to instantiate
                an instance of :attr:`~adorn.data.parameter.Parameter.cls`

        Returns:
            Optional[TypeCheckError]: If ``TypeCheckError``, there is an
                issue that would prevent ``obj`` from being converted to an
                instance of :attr:`~adorn.data.parameter.Parameter.cls`,
                otherwise ``None``
        """  # noqa: B950, RST304
        if (target_cls.cls == Any) or (
            getattr(target_cls.cls, "__origin__", None) is None
            and isinstance(obj, target_cls.cls)
        ):
            # object is already instantiated, so we assume it is ok
            return None
        return orchestrator.type_check(target_cls.cls, obj)


class Dependent(ParameterValue):
    """``Dependent`` allows for a parameter to leverage information from other parameters."""  # noqa: B950

    @staticmethod
    def get_constructor(target_cls: Parameter, obj: Dict[Any, Any]) -> Constructor:
        """Generate :class:`~adorn.data.constructor.Constructor` for relevant sub-type

        Args:
            target_cls (Parameter): container of the parent class
            obj (Dict[Any, Any]): arguments to instantiate
                :attr:`~adorn.data.parameter.Parameter.cls`

        Returns:
            Constructor: The constructor for the sub class specified
                 in ``obj``
        """  # noqa: RST304
        subcls = target_cls.args[0].resolve_class_name(obj["type"])
        return Constructor(subcls)

    @staticmethod
    def get_args(
        target_cls: Parameter,
        literal_dict: Dict[str, str],
        dependent_from_obj: bool = False,
        max_depth: Optional[int] = None,
    ) -> Dict[str, Any]:
        """Map the key in literal_dict to its requested dependent state

        Args:
            target_cls (Parameter): container of the local state
            literal_dict (Dict[str, str]): the request for dependent state
            dependent_from_obj (bool): if ``True``, the request is grabbing
                instantiated values, otherwise configured values
            max_depth (Optional[int]): the level at which local state will
                be inspected

        Returns:
            Dict[str, Any]: relevant local state needed by the parameter
        """
        # TODO: currently default args are not supported, this could
        # be accomplished by using the constructor of the object that
        # contains the needed state
        acc = dict()
        for k, v in literal_dict.items():
            local_state = target_cls.local_state
            keys = v.split(".")
            key_len = len(keys)
            upper_bound = (
                max_depth if max_depth is not None and dependent_from_obj else key_len
            )
            not_bad = True
            idx = 0
            while not_bad and idx < upper_bound:
                key = keys[idx]
                if idx == 0 or not dependent_from_obj:
                    # initial key or a DependentTypeCheck._type_check
                    if key not in local_state:
                        not_bad = False
                    else:
                        local_state = local_state[key]
                else:
                    if hasattr(local_state, key):
                        local_state = getattr(local_state, key)
                    else:
                        not_bad = False
                idx += 1
            if not_bad:
                acc[k] = local_state
        return acc

    @classmethod
    def check_args(
        cls, target_cls: Parameter, orchestrator: "Orchestrator"
    ) -> Optional[TypeCheckError]:
        """Ensure type level args are correct

        Exceptions:
            - :class:`~adorn.exception.type_check_error.MalformedDependencyError`:
                the ``Dependent`` type didn't specify the appropriate number of
                arguments
            - :class:`~adorn.exception.type_check_error.MissingLiteralError`:
                the first argument, zero based counting, wasn't a ``typing.Literal``
            - :class:`~adorn.exception.type_check_error.UnaryLiteralError`:
                ``typing.Literal`` specifies more than one argument
            - :class:`~adorn.exception.type_check_error.MalformedLiteralError`:
                ``typing.Literal`` wrapped a value that wasn't of type
                ``Dict[str, str]``

        Args:
            target_cls (Parameter): parameter that is dependent on
                state in the constructor
            orchestrator (Orchestrator): container of types, used
                to check the argument specified in the ``typing.Literal``

        Returns:
            Optional[TypeCheckError]: ``TypeCheckError`` specifying an
                error with the type level dependent argument, otherwise
                ``None``
        """
        if target_cls.args is None or len(target_cls.args) != 2:
            # need to specify a type and a mapping
            return MalformedDependencyError(target_cls=target_cls)

        potential_literal = target_cls.args[1]
        literal_origin = getattr(potential_literal, "__origin__", None)
        if literal_origin != Literal:
            # first arg must be wrapped in a literal
            return MissingLiteralError(target_cls=target_cls)

        literal_arg = getattr(potential_literal, "__args__", None)
        if literal_arg is None or len(literal_arg) != 1:
            # may only provide a single literal arg
            return UnaryLiteralError(target_cls=target_cls)

        literal_arg = literal_arg[0]

        literal_arg_check = orchestrator.type_check(Dict[str, str], literal_arg)
        if literal_arg_check is not None:
            # arg to a literal must be Dict[str, str]
            return MalformedLiteralError(
                target_cls=target_cls,
                literal_type=Dict[str, str],
                child=literal_arg_check,
            )

    @classmethod
    def check_literal_dict(
        cls,
        target_cls: Parameter,
        obj: Dict[Any, Any],
        dependent_from_obj: bool = False,
    ) -> Optional[TypeCheckError]:
        """Ensure literal dict is logical given the class and local state

        Exceptions:
            - :class:`~adorn.exception.type_check_error.ExtraLiteralError`:
                ``typing.Literal`` requested state that weren't part of the
                constructor of the parameter
            - :class:`~adorn.exception.type_check_error.TooDeepLiteralError`:
                ``typing.Literal`` requested state that was more than one layer deep.

        Args:
            target_cls (Parameter): container of the local state and dependency request
            obj (Dict[Any, Any]): arguments for the parameter
            dependent_from_obj (bool): if ``True``, the request is grabbing
                instantiated values, otherwise configured values

        Returns:
            Optional[TypeCheckError]: ``TypeCheckError`` if the request for state
                didn't work with the state that existed, otherwise ``None``
        """
        constructor = cls.get_constructor(target_cls=target_cls, obj=obj)
        literal_dict = target_cls.args[1].__args__[0]
        # check key is actually a parameter to the given class's constructor
        additional_keys = [
            i for i in literal_dict.keys() if i not in constructor.parameters
        ]
        if additional_keys:
            return ExtraLiteralError(target_cls, additional_keys)

        # check values, max, go one layer deep
        split_values = [(i.split("."), i) for i in literal_dict.values()]
        too_deep = [j for i, j in split_values if len(i) > 2]
        if too_deep:
            return TooDeepLiteralError(target_cls, too_deep)

        # check that value exists in the local state
        args = cls.get_args(
            target_cls=target_cls,
            literal_dict=literal_dict,
            dependent_from_obj=dependent_from_obj,
            max_depth=1 if dependent_from_obj else None,
        )
        literal_dict_key_check = cls.check_literal_dict_keys(
            target_cls=target_cls, literal_dict=literal_dict, args=args
        )
        if literal_dict_key_check is not None:
            return literal_dict_key_check

    @classmethod
    def check_literal_dict_keys(
        cls, target_cls: Parameter, literal_dict: Dict[str, str], args: Dict[str, Any]
    ) -> Optional[TypeCheckError]:
        """Ensure all requested args exist in local state

        Exceptions:
            - :class:`~adorn.exception.type_check_error.MissingDependencyError`:
                ``typing.Literal`` requested state that wasn't specified in the
                local state

        Args:
            target_cls (Parameter): container parameter's state
            literal_dict (Dict[str, str]): dependency request
            args (Dict[str, Any]): local state

        Returns:
            Optional[TypeCheckError]: ``TypeCheckError`` if the dependency
                request can't be fulfilled by the local state, otherwise ``None``
        """
        missing_keys = literal_dict.keys() - args.keys()
        if missing_keys:
            return MissingDependencyError(
                target_cls=target_cls,
                missing_dependency={k: literal_dict[k] for k in missing_keys},
            )

    @classmethod
    def perfunctory_type_check(
        cls,
        target_cls: Parameter,
        orchestrator: "Orchestrator",
        obj: Dict[Any, Any],
        dependent_from_obj: bool = False,
    ) -> Optional[TypeCheckError]:
        """Ensure the Dependent type specified is well formed

        Exceptions:
            - see :class:`~adorn.unit.parameter_value.Dependent.check_args`
            - see :class:`~adorn.unit.complex.Complex.general_check`
            - see :class:`~adorn.unit.parameter_value.Dependent.check_literal_dict`

        Args:
            target_cls (Parameter): specification of the parameter
                and its associated state
            orchestrator (Orchestrator): container of types, used
                to check nested types
            obj (Dict[Any, Any]): the arguments to construct the parameter
            dependent_from_obj (bool): if ``True``, the request is grabbing
                instantiated values, otherwise configured values

        Returns:
            Optional[TypeCheckError]: ``TypeCheckError`` if there was an issue
                with the ``DependentType``, otherwise ``None``
        """
        check_arg_ouptut = cls.check_args(
            target_cls=target_cls, orchestrator=orchestrator
        )
        if check_arg_ouptut is not None:
            return check_arg_ouptut

        # perform a Complex.general_check on the given wrapped type
        wrapped_cls = target_cls.args[0]
        cls_unit = orchestrator.get(wrapped_cls)
        wrapped_cls_general_check = cls_unit.general_check(
            wrapped_cls, orchestrator, obj
        )
        if wrapped_cls_general_check is not None:
            return wrapped_cls_general_check

        # check that the literal dict is well specified
        check_literal_dict_output = cls.check_literal_dict(
            target_cls=target_cls, obj=obj, dependent_from_obj=dependent_from_obj
        )
        if check_literal_dict_output is not None:
            return check_literal_dict_output

    @classmethod
    def _from_obj(
        cls, target_cls: Parameter, orchestrator: "Orchestrator", obj: Dict[Any, Any]
    ) -> Dict[Any, Any]:
        """Instantiate a parameter, using arguments and state from the :class:`~adorn.data.constructor.Constructor`

        Exceptions:
            - :class:`~adorn.exception.type_check_error.MalformedDependencyError`:
                the ``Dependent`` type didn't specify the appropriate number of
                arguments
            - :class:`~adorn.exception.type_check_error.MissingLiteralError`:
                the first argument, zero based counting, wasn't a ``typing.Literal``
            - :class:`~adorn.exception.type_check_error.UnaryLiteralError`:
                ``typing.Literal`` specifies more than one argument
            - :class:`~adorn.exception.type_check_error.MalformedLiteralError`:
                ``typing.Literal`` wrapped a value that wasn't of type
                ``Dict[str, str]``

        Args:
            target_cls (Parameter): the requested type and relevant state
            orchestrator (Orchestrator): container of all types, typically used
                to recurse down nested types
            obj (Any): an instance, containing the arguments for
                :attr:`~adorn.data.parameter.Parameter.cls`

        Returns:
            Any: An instance of :attr:`~adorn.data.parameter.Parameter.cls`
        """  # noqa: B950, DAR401, RST304
        check_arg_ouptut = cls.check_args(
            target_cls=target_cls, orchestrator=orchestrator
        )
        if check_arg_ouptut is not None:
            raise check_arg_ouptut
        literal_dict = target_cls.args[1].__args__[0]

        args = cls.get_args(
            target_cls=target_cls, literal_dict=literal_dict, dependent_from_obj=True
        )

        # check all required state exists
        literal_dict_key_check = cls.check_literal_dict_keys(
            target_cls=target_cls, literal_dict=literal_dict, args=args
        )
        if literal_dict_key_check is not None:
            raise literal_dict_key_check

        # enrich obj with dependent state
        for k, v in args.items():
            if k not in obj:
                obj[k] = v
        subcls = target_cls.args[0].resolve_class_name(obj["type"])
        return orchestrator.from_obj(subcls, obj)


@ParameterValue.register("dependent_type_check")
class DependentTypeCheck(Generic[_T, DictStrStr], Dependent):
    """Parameter that requires local state that is known before instantiation

    The first argument must be a :class:`~adorn.unit.complex.Complex`, that requires
    state, known before construction, from the constructor.  The second argument is
    a ``typing.Literal[Dict[str, str]]`` where the keys specifies the names of the
    dependent parameters and the values are the state needed from the constructor.
    """

    @classmethod
    def _contains(cls, obj: Parameter, orchestrator: "Orchestrator") -> bool:
        """Check if the ``obj`` is an instance of ``DependentTypeCheck``

        Args:
            obj (Parameter): potentially an instance of
                ``DependentTypeCheck``
            orchestrator (Orchestrator): container of all types, typically
                used to recurse nested types

        Returns:
            bool: if ``True``, ``obj`` is an instance of
                ``DependentTypeCheck``
        """  # noqa: B950
        if obj.origin != DependentTypeCheck:
            return False

        return (orchestrator.contains(obj.args[0])) and issubclass(obj.args[0], Complex)

    @classmethod
    def _type_check(
        cls, target_cls: Parameter, orchestrator: "Orchestrator", obj: Dict[Any, Any]
    ) -> Optional[TypeCheckError]:
        """Type check a parameter, using arguments and state from the :class:`~adorn.data.constructor.Constructor`

        Exceptions:
            - see :class:`~adorn.unit.parameter_value.Dependent.perfunctory_type_check`
            - :class:`~adorn.exception.type_check_error.KeyValueError`:
                errors for the argument(s) from ``obj`` and/or local state

        Args:
            target_cls (Parameter): the requested type and relevant state
            orchestrator (Orchestrator): container of all types, typically used
                to recurse down nested types
            obj (Any): an instance, containing the arguments for
                :attr:`~adorn.data.parameter.Parameter.cls`

        Returns:
            Optional[TypeCheckError]: ``TypeCheckError``, if the state and arguments
                can't generate an instance of :attr:`~adorn.data.parameter.Parameter.cls`,
                otherwise ``None``
        """  # noqa: B950, RST304
        perfunctory_type_check_output = cls.perfunctory_type_check(
            target_cls=target_cls, orchestrator=orchestrator, obj=obj
        )
        if perfunctory_type_check_output is not None:
            return perfunctory_type_check_output

        literal_dict = target_cls.args[1].__args__[0]

        constructor = cls.get_constructor(target_cls=target_cls, obj=obj)
        # perform a Complex.general_check on the requested values
        # need to parse the correct Complex instance for the given
        # parameter's type
        args = Params(cls.get_args(target_cls=target_cls, literal_dict=literal_dict))
        bad_args = dict()
        for k, v in args.items():
            sub = constructor.parameters[k].annotation
            unit = orchestrator.get(sub)
            arg_type_check = None
            if getattr(sub, "__origin__", None) is None and isinstance(v, sub):
                # object is already instantiated, so we assume it is ok
                arg_type_check = None
            elif isinstance(unit, Complex):
                # we do a shallow check on complex types because may contain
                # their own dependencies.
                arg_type_check = unit.general_check(
                    sub, orchestrator=orchestrator, obj=v
                )
            else:
                arg_type_check = orchestrator.type_check(sub, v)
            if arg_type_check is not None:
                bad_args[k] = arg_type_check
        if bad_args:
            return KeyValueError(target_cls, bad_args, obj)

        # generate an instance of Constructor where the key's in the literal_dict
        # are removed from the parameters attribute
        for k in literal_dict:
            if k not in obj:
                constructor.parameters.pop(k)
        return orchestrator.type_check(constructor, obj)


@ParameterValue.register("dependent_from_obj")
class DependentFromObj(Generic[_T, DictStrStr], Dependent):
    """Parameter that requires local state that is known after instantiation

    The first argument must be a :class:`~adorn.unit.complex.Complex`, that requires
    state, known after construction, from the constructor.  The second argument is
    a ``typing.Literal[Dict[str, str]]`` where the keys specifies the names of the
    dependent parameters and the values are the state needed from the constructor.
    """

    @classmethod
    def _contains(cls, obj: Parameter, orchestrator: "Orchestrator") -> bool:
        """Check if the ``obj`` is an instance of ``DependentFromObj``

        Args:
            obj (Parameter): potentially an instance of
                ``DependentFromObj``
            orchestrator (Orchestrator): container of all types, typically
                used to recurse nested types

        Returns:
            bool: if ``True``, ``obj`` is an instance of
                ``DependentFromObj``
        """
        if obj.origin != DependentFromObj:
            return False

        return (orchestrator.contains(obj.args[0])) and issubclass(obj.args[0], Complex)

    @classmethod
    def _type_check(
        cls, target_cls: Parameter, orchestrator: "Orchestrator", obj: Dict[Any, Any]
    ) -> Optional[TypeCheckError]:
        """Type check a parameter, using arguments from ``obj``

        Any local state that has been specified in the second argument to
        ``DependentFromObj`` will not be checked, because nothing has been
        instantiated, yet.

        Exceptions:
            - see :class:`~adorn.unit.parameter_value.Dependent.perfunctory_type_check`

        Args:
            target_cls (Parameter): the requested type and relevant state
            orchestrator (Orchestrator): container of all types, typically used
                to recurse down nested types
            obj (Any): an instance, containing the arguments for
                :attr:`~adorn.data.parameter.Parameter.cls`

        Returns:
            Optional[TypeCheckError]: ``TypeCheckError``, if the state and arguments
                can't generate an instance of :attr:`~adorn.data.parameter.Parameter.cls`,
                otherwise ``None``
        """  # noqa: B950, RST30
        perfunctory_type_check_output = cls.perfunctory_type_check(
            target_cls=target_cls,
            orchestrator=orchestrator,
            obj=obj,
            dependent_from_obj=True,
        )
        if perfunctory_type_check_output is not None:
            return perfunctory_type_check_output

        literal_dict = target_cls.args[1].__args__[0]
        constructor = cls.get_constructor(target_cls=target_cls, obj=obj)

        # generate an instance of Constructor where the key's in the literal_dict
        # are removed from the parameters attribute
        for k in literal_dict:
            if k not in obj:
                constructor.parameters.pop(k)

        return orchestrator.type_check(constructor, obj)


_DependentT = TypeVar("_DependentT", bound=Dependent)


@ParameterValue.register("dependent_union")
class DependentUnion(Generic[_DependentT, _T], Dependent):
    """Similar to ``typing.Union``, but only accepts two arguments, where the zeroth argument must be :class:`~adorn.unit.parameter_value.Dependent`"""  # noqa: B950

    @staticmethod
    def get_parameter_subtype(target_cls: Parameter, arg: Type) -> Parameter:
        """Generate an Parameter where we update the ``cls`` arg with a type option.

        Args:
            target_cls (Parameter): state and ``DependentUnion`` request
            arg (Type): a type option specified within ``DependentUnion``

        Returns:
            Parameter: state with :attr:`~adorn.data.parameter.Parameter.cls`
                argument specifying a type option specified in ``DependentUnion``
        """  # noqa: RST304
        return Parameter(
            cls=arg,
            parent=target_cls.parent,
            local_state=target_cls.local_state,
            parameter_name=target_cls.parameter_name,
        )

    @classmethod
    def _contains(cls, obj: Parameter, orchestrator: "Orchestrator") -> bool:
        """Check if the ``obj`` is an instance of ``DependentUnion``

        Args:
            obj (Parameter): potentially an instance of
                ``DependentUnion``
            orchestrator (Orchestrator): container of all types, typically
                used to recurse nested types

        Returns:
            bool: if ``True``, ``obj`` is an instance of
                ``DependentUnion``
        """
        if obj.origin != DependentUnion:
            return False
        origin = getattr(obj.args[0], "__origin__", None)
        if origin is None or not issubclass(origin, Dependent):
            return False
        return all(
            orchestrator.contains(
                cls.get_parameter_subtype(target_cls=obj, arg=obj.args[i])
            )
            for i in range(2)
        )

    @classmethod
    def get_parameter(
        cls, target_cls: Parameter, orchestrator: "Orchestrator", obj: Dict[Any, Any]
    ) -> Union[Parameter, KeyValueError]:
        """Type Check the given ``obj`` against all potential types

        Args:
            target_cls (Parameter): state and list of potential types ``obj``
                could be
            orchestrator (Orchestrator): container of all types, typically
                used to recurse nested types
            obj (Dict[Any, Any]): arguments for one of the listed types

        Returns:
            Union[Parameter, KeyValueError]: either a ``Parameter`` which wraps the type
                ``obj`` can be coerced into, otherwise ``KeyValueError`` containing the
                reasons why ``obj`` cannot be coerced into any of the listed types
        """
        key_values = []
        for i in target_cls.args:
            nu_parameter = cls.get_parameter_subtype(target_cls=target_cls, arg=i)
            type_check_output = orchestrator.type_check(nu_parameter, obj)
            if type_check_output is None:
                return nu_parameter
            key_values.append((nu_parameter, type_check_output))
        return KeyValueError(target_cls=target_cls, key_values=key_values, obj=obj)

    @classmethod
    def _from_obj(
        cls, target_cls: Parameter, orchestrator: "Orchestrator", obj: Dict[Any, Any]
    ) -> Dict[Any, Any]:
        """Convert ``obj`` into one of the listed types

        Args:
            target_cls (Parameter): state and list of potential types ``obj``
                could be
            orchestrator (Orchestrator): container of all types, typically
                used to recurse nested types
            obj (Dict[Any, Any]): arguments for one of the listed types

        Returns:
            Any: an instance of one of the listed types

        Raises:
            KeyValueError: ``obj`` can't be coerced into any of the listed types
        """  # noqa: DAR401, DAR402
        output = cls.get_parameter(
            target_cls=target_cls, orchestrator=orchestrator, obj=obj
        )
        if isinstance(output, KeyValueError):
            raise output
        return orchestrator.from_obj(output, obj)

    @classmethod
    def _type_check(
        cls, target_cls: Parameter, orchestrator: "Orchestrator", obj: Dict[Any, Any]
    ) -> Optional[KeyValueError]:
        """Type check ``obj`` against all potential types

        Args:
            target_cls (Parameter): state and list of potential types ``obj``
                could be
            orchestrator (Orchestrator): container of all types, typically
                used to recurse nested types
            obj (Dict[Any, Any]): arguments for one of the listed types

        Returns:
            Optional[KeyValueError]: if ``obj`` can't be coerced into any of the type,
                a ``KeyValueError`` is returned containing the reasons why ``obj``
                cannot be coerced into any of the listed types, otherwise ``None``
        """
        output = cls.get_parameter(
            target_cls=target_cls, orchestrator=orchestrator, obj=obj
        )
        if isinstance(output, KeyValueError):
            return output
        return None



# Copyright 2021 Jacob Baumbach
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Different ways to handle instances of :class:`~adorn.data.constructor.Constructor`"""
import inspect
from collections import defaultdict
from typing import Any
from typing import Callable
from typing import Dict
from typing import Optional
from typing import Set
from typing import Type
from typing import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
    from adorn.orchestrator.orchestrator import Orchestrator

from adorn.data.parameter import Parameter
from adorn.data.constructor import Constructor
from adorn.exception.type_check_error import (
    KeyValueDiffError,
    KeyValueError,
    ParameterOrderError,
    TypeCheckError,
    WrongTypeError,
)
from adorn.unit.simple import Simple


class ConstructorValue(Simple):
    """:class:`~adorn.unit.unit.Unit` that contains all the ways to handle :class:`~adorn.data.constructor.Constructor` instances."""  # noqa: B950

    _registry = defaultdict(dict)

    def contains(self, obj: Any, orchestrator: "Orchestrator") -> bool:
        """Check if the ``obj`` is an instance of :class:`~adorn.data.constructor.Constructor`

        Args:
            obj (Any): potentially an instance of
                :class:`~adorn.data.constructor.Constructor`
            orchestrator (Orchestrator): this is unused

        Returns:
            bool: if ``True``, ``obj`` is an instance of
                :class:`~adorn.data.constructor.Constructor`
        """  # noqa: B950
        return isinstance(obj, Constructor)

    def get(
        self, key: Any, orchestrator: "Orchestrator"
    ) -> Optional["ConstructorValue"]:
        """Get the relevant class to handle the given :class:`~adorn.data.constructor.Constructor` instance.

        The :attr:`~adorn.unit.complex.Complex.constructor_type` attached to the
        :attr:`~adorn.data.constructor.Constructor.subclass` is used to determine which
        subclass of ``ConstructorValue`` to use.  If
        :attr:`~adorn.unit.complex.Complex.constructor_type` is ``None`` then
        :class:`~adorn.unit.constructor_value.BaseConstructorValue` will be returned

        Args:
            key (Any): potentially an instance of
                :class:`~adorn.data.constructor.Constructor``
            orchestrator (Orchestrator): container of all types, typically
                used to recurse nested types

        Returns:
            Optional[ConstructorValue]: if ``key`` is an instance of
                :class:`~adorn.data.constructor.Constructor` then a subclass of
                ``ConstructorValue`` will be returned, otherwise ``None``
        """  # noqa: B950, RST304
        if not self.contains(key, orchestrator):
            return None
        return self._registry[ConstructorValue].get(key.constructor_type or "base")

    def from_obj(
        self, target_cls: Constructor, orchestrator: "Orchestrator", obj: Any
    ) -> Any:
        """Instantiate the arguments and an instance of :attr:`~adorn.data.constructor.Constructor.subclass`

        Args:
            target_cls (Constructor): instance of
                :class:`~adorn.data.constructor.Constructor`
            orchestrator (Orchestrator): container of all types, typically used
                to recurse down nested types
            obj (Any): an instance, containing the arguments for
                :attr:`~adorn.data.constructor.Constructor.subclass`

        Returns:
            Any: An instance of :attr:`~adorn.data.constructor.Constructor.subclass`
        """  # noqa: B950, RST304
        constructor_value = self.get(target_cls, orchestrator)
        return constructor_value._from_obj(target_cls, orchestrator, obj)

    def type_check(
        self, target_cls: Constructor, orchestrator: "Orchestrator", obj: Any
    ) -> Optional[TypeCheckError]:
        """Check the arguments for :attr:`~adorn.data.constructor.Constructor.subclass`

        Exceptions:
            - :class:`~adorn.exception.type_check_error.WrongTypeError`:
                ``target_cls`` was not an instance of
                :class:`~adorn.data.constructor.Constructor`


        Args:
            target_cls (Constructor): instance of
                :class:`~adorn.data.constructor.Constructor`
            orchestrator (Orchestrator): container of all types, typically used
                to recurse down nested types
            obj (Any): an instance, containing the arguments for
                :attr:`~adorn.data.constructor.Constructor.subclass`

        Returns:
            Optional[TypeCheckError]: if ``TypeCheckError`` then either
                ``target_cls`` was not an instance of
                :class:`~adorn.data.constructor.Constructor`
                or there are issues which would prevent creating the arguments
                for :attr:`~adorn.data.constructor.Constructor.subclass`,
                otherwise ``None``
        """  # noqa: RST304
        if not self.contains(target_cls, orchestrator):
            return WrongTypeError(Constructor, target_cls)

        constructor_value = self.get(target_cls, orchestrator)
        return constructor_value._type_check(target_cls, orchestrator, obj)

    @staticmethod
    def get_kv(
        main: Dict[str, Any],
        alter_value: Callable[..., type],
        sub: Dict[str, Any],
        missing: Set[str],
        check: Callable[..., bool] = lambda _: False,
    ) -> Optional[Dict[str, Type]]:
        """Generate the difference in keys between ``main`` and ``sub``.

        Args:
            main (Dict[str, Any]): a dict, whose keys are considered necessary
            alter_value (Callable[..., type]): an alteration performed to a value
                of ``main`` before it is added to the returned result
            sub (Dict[str, Any]): a dict which is checked that it contains the
                keys in ``main``
            missing (Set[str]): a collection of keys to be ignored in ``main``
            check (Callable[..., bool]): an additional check on the keys of ``main``,
                which if ``True`` will prevent the key from being checked

        Returns:
            Optional[Dict[str, Type]]: the keys and associated types that were missing
                in ``sub``.  If no keys were missing then ``None``
        """
        extras = dict()
        for k in main.keys():
            if k in missing or check(main[k]):
                continue
            if k not in sub:
                extras[k] = alter_value(main[k])
        if extras:
            return extras

    @staticmethod
    def parameter_consistency(
        target_cls: Constructor, obj: Dict[Any, Any]
    ) -> Optional[KeyValueDiffError]:
        """Check that ``obj`` has all the necessary arguments, and nothing more.

        Args:
            target_cls (Constructor): contains the parameters required by
                :attr:`~adorn.data.constructor.Constructor.subclass`
            obj (Dict[Any, Any]): collection of potential arguments for
                :attr:`~adorn.data.constructor.Constructor.subclass`

        Returns:
            Optional[KeyValueDiffError]: ``None`` if ``obj`` contained the correct
                set of keys, otherwise ``KeyValueDiffError`` is returned which specifies
                the missing and/or additional arguments in ``obj``
        """  # noqa: RST304
        parameter_kv = ConstructorValue.get_kv(
            target_cls.parameters,
            lambda i: i.annotation,
            obj,
            {"self"},
            lambda i: i.default != inspect._empty,
        )
        obj_kv = ConstructorValue.get_kv(obj, type, target_cls.parameters, {"type"})

        if parameter_kv or obj_kv:
            return KeyValueDiffError(
                target_cls=target_cls.subclass,
                parameter_kv=parameter_kv,
                obj_kv=obj_kv,
                obj=obj,
            )
        return None

    @staticmethod
    def dict_type_check(
        target_cls: Constructor,
        orchestrator: "Orchestrator",
        obj: Dict[Any, Any],
    ) -> Optional[KeyValueError]:
        """Type check all the arguments in ``obj``

        Args:
            target_cls (Constructor): instance of
                :class:`~adorn.data.constructor.Constructor`, which
                specifies the parameters for the constructor
            orchestrator (Orchestrator): container of all types, used
                to check the arguments
            obj (Any): an instance, containing the arguments for
                :attr:`~adorn.data.constructor.Constructor.subclass`

        Returns:
            Optional[KeyValueError]: if all the args type check then ``None``,
                otherwise ``KeyValueError``, which specifies the args that
                failed and why
        """  # noqa: RST304
        exception_dict = dict()
        for k in target_cls.parameter_order or obj.keys():
            if k in obj and (k not in {"type"}):
                parameter = Parameter(
                    target_cls.parameters[k].annotation, target_cls, obj, k
                )
                output = orchestrator.type_check(parameter, obj[k])
                if output is not None:
                    exception_dict[k] = output

        if exception_dict:
            return KeyValueError(
                target_cls=target_cls.subclass, key_values=exception_dict, obj=obj
            )
        return None

    @staticmethod
    def check_parameter_order(target_cls: Constructor) -> Optional[ParameterOrderError]:
        """Ensure :attr:`~adorn.unit.complex.Complex.parameter_order` specifies all the parameters.

        Args:
            target_cls (Constructor): :attr:`~adorn.data.constructor.Constructor.subclass`
                specifies a :attr:`~adorn.unit.complex.Complex.parameter_order`

        Returns:
            Optional[ParameterOrderError]: if :attr:`~adorn.unit.complex.Complex.parameter_order`
                is complete then ``None``, otherwise ``ParameterOrderError``, which specifies
                the additional and/or missing parameters in
                :attr:`~adorn.unit.complex.Complex.parameter_order`
        """  # noqa: B950, RST304
        # no parameter_order, then nothing to check
        if target_cls.parameter_order is None:
            return None

        po_set = set(target_cls.parameter_order)
        pk = target_cls.parameters.keys()
        too_few = pk - po_set
        too_many = po_set - pk
        if too_few or too_many:
            return ParameterOrderError(
                target_cls=target_cls, too_few=too_few, too_many=too_many
            )


@ConstructorValue.register("base")
class BaseConstructorValue(ConstructorValue):
    """Vanilla way of type checking and instantiating :class:`~adorn.data.constructor.Constructor` instances"""  # noqa: B950

    @classmethod
    def _from_obj(
        cls, target_cls: Constructor, orchestrator: "Orchestrator", obj: Dict[Any, Any]
    ) -> Dict[Any, Any]:
        """Instantiate the arguments and an instance of :attr:`~adorn.data.constructor.Constructor.subclass`

        Args:
            target_cls (Constructor): instance of
                :class:`~adorn.data.constructor.Constructor`
            orchestrator (Orchestrator): container of all types, typically used
                to recurse down nested types
            obj (Dict[Any, Any]): an instance, containing the arguments for
                :attr:`~adorn.data.constructor.Constructor.subclass`

        Returns:
            Any: An instance of :attr:`~adorn.data.constructor.Constructor.subclass`

        Raises:
            ParameterOrderError: specifies the additional and/or missing parameters
                in :attr:`~adorn.unit.complex.Complex.parameter_order`
        """  # noqa: B950, DAR401, DAR402, RST304
        type_str = obj.pop("type", None)
        paramterer_order_check = cls.check_parameter_order(target_cls=target_cls)
        if paramterer_order_check is not None:
            raise paramterer_order_check
        kwargs = dict()
        for k in target_cls.parameter_order or obj.keys():
            if k in obj:
                parameter = Parameter(
                    target_cls.parameters[k].annotation, target_cls, kwargs, k
                )
                kwargs[k] = orchestrator.from_obj(parameter, obj[k])
        obj["type"] = type_str
        return target_cls.constructor(**kwargs)

    @classmethod
    def _type_check(
        cls, target_cls: Constructor, orchestrator: "Orchestrator", obj: Dict[Any, Any]
    ) -> Optional[TypeCheckError]:
        """Check the arguments for :attr:`~adorn.data.constructor.Constructor.subclass`

        Exceptions:
            - :class:`~adorn.exception.type_check_error.ParameterOrderError`:
                specifies the additional and/or missing parameters
                in :attr:`~adorn.unit.complex.Complex.parameter_order`
            - :class:`~adorn.exception.type_check_error.KeyValueDiffError`:
                missing and/or additional arguments in ``obj``
            - :class:`~adorn.exception.type_check_error.KeyValueError`:
                args in ``obj`` didn't type check

        Args:
            target_cls (Constructor): instance of
                :class:`~adorn.data.constructor.Constructor`
            orchestrator (Orchestrator): container of all types, typically used
                to recurse down nested types
            obj (Dict[Any, Any]): an instance, containing the arguments for
                :attr:`~adorn.data.constructor.Constructor.subclass`

        Returns:
            Optional[TypeCheckError]: if ``TypeCheckError`` there are
                issues which would prevent creating the arguments
                for :attr:`~adorn.data.constructor.Constructor.subclass`,
                otherwise ``None``
        """  # noqa: RST304
        paramterer_order_check = cls.check_parameter_order(target_cls=target_cls)
        if paramterer_order_check is not None:
            return paramterer_order_check
        consistency_message = cls.parameter_consistency(target_cls, obj)
        if consistency_message:
            return consistency_message
        return cls.dict_type_check(
            target_cls=target_cls, orchestrator=orchestrator, obj=obj
        )



# Copyright 2021 Jacob Baumbach
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Abstract classes for creating a search space for an ``adorn`` config"""
from adorn.unit.complex import Complex


class SearchElement(Complex):
    """Specification of a single component to search over"""


class SearchSpace(Complex):
    """Specification of multiple components to search over"""

    def __len__(self) -> int:  # pragma: no cover
        """Number of elements to be searched over

        Raises:
            NotImplementedError: needs to be implemented by child class
        """
        raise NotImplementedError



# Copyright 2021 Jacob Baumbach
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Apply an alteration to an object based off a ``dict``"""
from os import environ
from typing import Any
from typing import Dict
from typing import Optional
from typing import Type
from typing import TYPE_CHECKING
from typing import Union

from adorn.exception.type_check_error import TypeCheckError
from adorn.exception.type_check_error import UserDictError

if TYPE_CHECKING:  # pragma: no cover
    from adorn.orchestrator.orchestrator import Orchestrator

from adorn.alter.alter import Alter
from adorn.data.parameter import Parameter
from adorn.params import Params


class UserDictAlter(Alter):
    """Inject a value from a user provided ``dict``

    Args:
        user_dict (Dict[str, Any]): dictionary which contains
            values to be injected
    """

    def __init__(self, user_dict: Dict[str, Any]) -> None:
        super().__init__()
        self.user_dict = user_dict

    def a_contains(
        self,
        target_cls: Type,
        orchestrator: "Orchestrator",
        obj: Any,
    ) -> bool:
        """Check if a value in ``user_dict`` has been requested

        Args:
            target_cls (Type): the target Type, the given ``obj`` will be converted
                into or checked against
            orchestrator (Orchestrator): container of all types, typically used
                to recurse down nested types
            obj (Any): an instance, that will be checked to see if it can be
                converted into the ``target_cls`` or will be converted to an
                instance of ``target_cls``

        Returns:
            bool: if ``True`` then a value in the user dict has been
                requested else ``False``
        """
        return (
            isinstance(target_cls, Parameter)
            and isinstance(obj, Params)
            and obj.get("type") == self.type_value
            and obj.get("key") in self.user_dict
        )

    def a_get(
        self,
        target_cls: Type,
        orchestrator: "Orchestrator",
        obj: Any,
    ) -> Optional[Alter]:
        """Return this ``Alter`` if the associated with the arguments are requesting it

        Args:
            target_cls (Type): the target Type, the given ``obj`` will be converted
                into or checked against
            orchestrator (Orchestrator): container of all types, typically used
                to recurse down nested types
            obj (Any): an instance, that will be checked to see if it can be
                converted into the ``target_cls`` or will be converted to an
                instance of ``target_cls``

        Returns:
            Optional[Alter]: if the provided arguments are contained by this ``Alter``
                return it, otherwise ``None``
        """
        if self.a_contains(
            target_cls=target_cls,
            orchestrator=orchestrator,
            obj=obj,
        ):
            return self

    def alter_obj(
        self,
        target_cls: Type,
        orchestrator: "Orchestrator",
        obj: Any,
    ) -> Union[Any, TypeCheckError]:
        """Swap ``obj`` with a value in ``user_dict``

        Args:
            target_cls (Type): the target Type, the given ``obj`` will be converted
                into or checked against
            orchestrator (Orchestrator): container of all types, typically used
                to recurse down nested types
            obj (Any): an instance, that will be checked to see if it can be
                converted into the ``target_cls`` or will be converted to an
                instance of ``target_cls``

        Returns:
            Union[Any, TypeCheckError]: a value from ``user_dict`` or a
                ``TypeCheckError`` providing a reason why a value from
                ``user_dict`` couldn't be provided
        """
        user_dict_value = self.user_dict[obj.get("key")]
        try:
            new_obj = orchestrator.from_obj(target_cls.cls, user_dict_value)
        except Exception as e:
            return UserDictError(type(self), target_cls, obj, e)
        return new_obj

    def log_instance(
        self,
        target_cls: Type,
        orchestrator: "Orchestrator",
        obj: Any,
        instance: Any,
    ) -> None:  # pragma: no cover
        """Apply logging to an instantiate object

        Args:
            target_cls (Type): the target Type, the given ``obj`` was converted
                into
            orchestrator (Orchestrator): container of all types, typically used
                to recurse down nested types
            obj (Any): an instance, that was converted to an
                instance of ``target_cls``
            instance (Any): the output of converting ``obj`` into an instance
                of ``target_cls``

        Returns:
            None: nothing is returned, this is a logging event
        """
        return None

    @property
    def type_value(self) -> str:
        """The expected value for the key ``type`` in the given :class:`~adorn.params.Params` object

        Returns:
            str: the value of ``type`` for which this ``Alter`` is appropriate
        """  # noqa: B950
        return "user_dict"


class EnvDictAlter(UserDictAlter):
    """Inject values from the environment"""

    def __init__(self) -> None:
        super().__init__(user_dict=environ)

    @property
    def type_value(self) -> str:
        """The expected value for the key ``type`` in the given :class:`~adorn.params.Params` object

        Returns:
            str: the value of ``type`` for which this ``Alter`` is appropriate
        """  # noqa: B950
        return "ENV"



# Copyright 2021 Jacob Baumbach
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Exceptions specifying why a request was malformed"""
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Set
from typing import Tuple
from typing import Type
from typing import TYPE_CHECKING
from typing import Union

if TYPE_CHECKING:  # pragma: no cover
    from adorn.data.parameter import Parameter
    from adorn.params import Params
    from adorn.unit.anum import Anum
    from adorn.unit.unit import Unit


DictOrList = Union[
    Dict[Union[int, str], "TypeCheckError"],
    List[Tuple[Dict[Union[int, str], "TypeCheckError"]]],
]


class TypeCheckError(Exception):
    """Generic ``Exception`` that contains the state the produced the error.

    Args:
        target_cls (Type): the requested type to perform an action against
        msg (List[str]): explanation about the exception
        child (Optional[TypeCheckError]): an excpetion produced by a downstream
            process
        obj (Optional[Any]): the caller provided object, that was innvolved in
            causing the error
    """

    def __init__(
        self,
        target_cls: Type,
        msg: List[str],
        child: Optional["TypeCheckError"] = None,
        obj: Optional[Any] = None,
    ) -> None:
        super().__init__()
        self.target_cls = target_cls
        self.msg = msg
        self.child = child
        self.obj = obj

    def to_str(self, seed: str = "") -> str:
        """Create a sting with the reason/explanation of the error.

        Args:
            seed (str): spacing used to compose error statements in a cute way

        Returns:
            str: information about the error
        """
        prefix = "\n" + seed if seed else ""
        line_break = "\n" + seed
        outer_msg = prefix + line_break.join(self.msg)
        if self.child is None:
            return outer_msg
        return outer_msg + self.child.to_str(seed + "\t")

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        return self.to_str()

    def __eq__(self, other) -> bool:
        if type(self) != type(other):
            return False
        return all(
            [
                self.target_cls == other.target_cls,
                self.msg == other.msg,
                self.child == other.child,
                self.obj == other.obj,
            ]
        )


class WrongTypeError(TypeCheckError):
    """Error when provided an ``obj`` of the wrong type

    Args:
        target_cls (type): the requested type to perform an action against
        obj (Any): caller provided object, that is not of type
            ``target_cls``
    """

    def __init__(self, target_cls: type, obj: Any) -> "TypeCheckError":
        target_cls_name = (
            target_cls.__name__ if hasattr(target_cls, "__name__") else target_cls
        )
        msg = [
            f"Expected an object of type {target_cls_name},\n ",
            "but received an object of type ",
            f"{type(obj).__name__}\nwith a value of:",
            f"\n\t{obj}",
        ]
        super().__init__(target_cls=target_cls, msg=msg, child=None, obj=obj)


class ParamError(TypeCheckError):
    """Error when an obj of not type :class:`~adorn.params.Params` is passed

    Args:
        target_cls (type): the requested type to perform an action against
        obj (Any): caller provided object, that is not of type
            ``Params``
    """

    def __init__(self, target_cls: type, obj: Any) -> "TypeCheckError":
        target_cls_name = (
            target_cls.__name__ if hasattr(target_cls, "__name__") else target_cls
        )
        msg = [
            f"For {target_cls_name}, expected a Params object",
            f"but received an object of type {type(obj).__name__}",
        ]
        super().__init__(target_cls=target_cls, msg=msg, child=None, obj=obj)


class UnRepresentedTypeError(TypeCheckError):
    """Error when a :class:`~adorn.unit.unit.Unit` does not cotain a given type.

    Args:
        target_cls (type): The requested type that can't be expressed by the ``Unit``
        unit (Unit): a collection of types that can't express ``target_cls``
        obj (Any): the caller created object that was supposed to map to ``target_cls``
    """

    def __init__(self, target_cls: type, unit: "Unit", obj: Any) -> "TypeCheckError":
        target_cls_name = (
            target_cls.__name__ if hasattr(target_cls, "__name__") else target_cls
        )
        msg = [
            (
                f"Requested type: {target_cls_name} with an object of type:"
                f"{type(obj).__name__}"
            ),
            "with a value of:",
            f"{obj}",
            f"didn't match any subtype of {type(unit).__name__}",
        ]
        super().__init__(target_cls=target_cls, msg=msg, child=None, obj=obj)


class KeyValueError(TypeCheckError):
    """Error when dependencies of the given ``target_cls`` produce an error.

    Args:
        target_cls (type): the requested type, whose dependencies produced error(s)
        key_values (DictOrList): a collection of the name of the dependency and
            error the dependency produced
        obj (Union[List[Any], Dict[str, Any]]): caller specified values for the
            dependencies
    """

    def __init__(
        self,
        target_cls: type,
        key_values: DictOrList,
        obj: Union[List[Any], Dict[str, Any]],
    ) -> None:
        msg = self.get_msg(target_cls, key_values)
        self.key_values = key_values
        super().__init__(target_cls, msg, child=None, obj=obj)

    @staticmethod
    def get_msg(
        target_cls: type,
        key_values: DictOrList,
    ) -> List[str]:
        """Generate the msg containing why the dependencies produced an error.

        Args:
            target_cls (type): the requested type, whose dependencies produced
                error(s)
            key_values (DictOrList): a collection of the name of the dependency
                and error the dependency produced

        Returns:
            List[str]: message containing the reasons why the dependencies produced
                errors
        """
        result_list = []
        collection = key_values.items() if isinstance(key_values, dict) else key_values
        for k, v in collection:
            msg = "\t-" + f"{k}: " + v.to_str("\t\t")
            result_list.append(msg)
        return [
            f"Failed to construct {target_cls} because of the following values:"
        ] + result_list


class HashableError(TypeCheckError):
    """Error when a Unit's subtype is not hashable.

    This is typically caused when a type is wrapped in a ``Set``, but the
    wrapped type is not hashable.

    Args:
        target_cls (Type): The requested type, that isn't hashable
        unit (Unit): the collection of types that contained ``target_cls``
        obj (Optional[Any]): the caller created object that maps
            to the unhashable type
    """

    def __init__(self, target_cls: Type, unit: "Unit", obj: Optional[Any]) -> None:
        target_origin = getattr(target_cls, "__origin__", None)
        target_arg = getattr(target_cls, "__args__", [None])[0]
        ta = target_arg or target_origin or target_cls
        msg = [
            f"Failed to construct {target_cls}, because {ta}",
            "is not hashable.  You may need to alter the hashable attr of",
            f"{unit.__name__} or use a type that is hashable",
        ]
        super().__init__(target_cls, msg, child=None, obj=obj)


class TupleArgLenError(TypeCheckError):
    """Exception when the number of args specified in the Tuple does not match the object's length.

    Args:
        target_cls (Type): the tuple type
        obj (Optional[Any]): the caller created object that has a length
            different than that specified by ``target_cls``
    """  # noqa: B950

    def __init__(self, target_cls: Type, obj: Optional[Any]) -> None:
        target_args = getattr(target_cls, "__args__", [None])
        msg = [
            f"Failed to create a {target_cls} because {len(target_args)} args were",
            f"expected but only {len(obj)} were received.",
            f"obj had a value of {obj}",
        ]
        super().__init__(target_cls, msg, child=None, obj=obj)


class KeyValueDiffError(TypeCheckError):
    """:class:`~adorn.params.Params` object had missing and/or additional information

    Args:
        target_cls (type): the type that the ``obj`` was intended to create
        parameter_kv (Optional[Dict[str, type]]): missing information from the ``obj``
        obj_kv (Optional[Dict[str, type]]): additional information in the ``obj``
        obj (Params): the ``Params`` object that has missing and/or additional
            information
    """

    def __init__(
        self,
        target_cls: type,
        parameter_kv: Optional[Dict[str, type]],
        obj_kv: Optional[Dict[str, type]],
        obj: "Params",
    ) -> None:
        msg = [
            f"{target_cls} was missing arguments, indicated with a `-`",
            "and/or passed additional arguments indicated with a `+`",
        ]
        if parameter_kv:
            missing = [
                f"\t- {k}: {getattr(v, '__name__', v)}" for k, v in parameter_kv.items()
            ]
            msg.extend(missing)

        if obj_kv:
            added = [f"\t+ {k}: {getattr(v, '__name__', v)}" for k, v in obj_kv.items()]
            msg.extend(added)
        self.parameter_kv = parameter_kv
        self.obj_kv = obj_kv
        super().__init__(target_cls, msg, child=None, obj=obj)


class ComplexTypeMismatchError(TypeCheckError):
    """The `type` specified doesn't match any of the options for the ``target_cls``

    Args:
        target_cls (Type): the requested type
        possibilities (Dict[str, type]): the different subclasses that can be
            requested for ``target_cls``
        obj (Params): ``Params`` with a value for ``type`` that doesn't map to
            any of the subclasses of ``target_cls``
    """

    def __init__(
        self, target_cls: Type, possibilities: Dict[str, type], obj: "Params"
    ) -> None:
        msg = [
            f"For {target_cls}, a Params object specified a `type`",
            f"of {obj.get('type')}, which is not an acceptable option",
            f"for {target_cls}.  The potential values for `type` and",
            "their corresponding type are:",
            *[f"\t- {k}: {v.__name__}" for k, v in possibilities.items()],
        ]
        self.possiblilites = possibilities
        super().__init__(target_cls, msg, child=None, obj=obj)


class ParameterOrderError(TypeCheckError):
    """The ``parameter_order`` attr contained too many and/or too few parameters

    Args:
        target_cls (type): the type with a malformed ``parameter_order`` attr
        too_few (Set[str]): parameters not in ``parameter_order``
        too_many (Set[str]): additional parameters in ``parameter_order`` that aren't
            in the constructor
    """

    def __init__(
        self,
        target_cls: type,
        too_few: Set[str],
        too_many: Set[str],
    ) -> None:
        msg = [
            f"{target_cls} parameter_order attribute, was missing parameters,",
            "indicated with a `-` and/or passed additional arguments",
            "indicated with a `+`",
        ]
        if too_few:
            missing = [f"\t- {k}" for k in too_few]
            msg.extend(missing)

        if too_many:
            added = [f"\t+ {k}" for k in too_many]
            msg.extend(added)

        self.too_few = too_few
        self.too_many = too_many
        super().__init__(target_cls, msg, child=None, obj=None)


class MalformedDependencyError(TypeCheckError):
    """:class:`~adorn.unit.parameter_value.Dependent` type was not provided all the necessary information.

    Args:
        target_cls (Type): The type with malformed dependency information
    """  # noqa: B950

    def __init__(self, target_cls: Type) -> None:
        msg = [
            f"{target_cls.cls} type was not provided all the necessary",
            "information at the type level.  Ensure that all necessary",
            "information was passed to the type.",
        ]
        super().__init__(target_cls, msg, child=None, obj=None)


class MissingLiteralError(TypeCheckError):
    """:class:`~adorn.unit.parameter_value.Dependent` type's first argument was not a ``typing.Literal``.

    Args:
        target_cls (Type): Type with a dependency that didn't have a ``typing.Literal``
    """  # noqa: B950

    def __init__(self, target_cls: Type) -> None:
        msg = [
            f"{target_cls.cls} requires it's last argument to be wrapped",
            "in a ``typing.Literal``",
        ]
        super().__init__(target_cls, msg, child=None, obj=None)


class UnaryLiteralError(TypeCheckError):
    """``typing.Literal`` that was passed more than one arg

    Args:
        target_cls (Type): Type with a dependency that had a ``typing.Literal``
            with more than one arg
    """

    def __init__(self, target_cls: Type) -> None:
        msg = [
            f"{target_cls.cls} requires it's last argument to be wrapped",
            "in a ``typing.Literal`` which is passed a single argument.",
        ]
        super().__init__(target_cls, msg, child=None, obj=None)


class MalformedLiteralError(TypeCheckError):
    """``typing.Literal`` argument wasn't of the correct type

    Args:
        target_cls (Type): the type which contained the problem ``typing.Literal``
        literal_type (Type): the required type of the ``typing.Literal``
            argument
        child (TypeCheckError): the error produced by the ``typing.Literal``
            argument
    """

    def __init__(
        self, target_cls: Type, literal_type: Type, child: TypeCheckError
    ) -> None:
        msg = [
            f"{target_cls.cls} requires it's last argument to be wrapped",
            f"in a ``typing.Literal`` which is of type {literal_type}",
        ]
        self.literal_type = literal_type
        super().__init__(target_cls, msg, child=child, obj=None)


class ExtraLiteralError(TypeCheckError):
    """``typing.Literal`` contains additional keys not specified in the constructor.

    Args:
        target_cls (Type): the type which contains the problem ``typing.Literal``
        extras (List[str]): the additional arguments specified in the
            ``typing.Literal`` argument
    """

    def __init__(self, target_cls: Type, extras: List[str]) -> None:
        msg = [
            f"{target_cls.cls}'s' ``typing.Literal`` arg contains additional keys",
            "that were not in the constructor",
            *[f"\t- {k}" for k in extras],
        ]
        self.extras = extras
        super().__init__(target_cls, msg, child=None, obj=None)


class TooDeepLiteralError(TypeCheckError):
    """``typing.Literal`` specified a dependency that goes beyond one layer deep

    An example would be `"a.b.c"`.

    Args:
        target_cls (Type): the type which contains the problem ``typing.Literal``
        bad_literal (List[str]): the dependency requests that went too deep
    """

    def __init__(self, target_cls: Type, bad_literal: List[str]) -> None:
        msg = [
            f"{target_cls.cls}'s' ``typing.Literal`` arg is not allowed to have",
            "dependencies that go beyond one layer deep, one period.",
            "The following dependencies were more than one layer deep:",
            *[f"\t- {k}" for k in bad_literal],
        ]
        self.bad_literal = bad_literal
        super().__init__(target_cls, msg, child=None, obj=None)


class MissingDependencyError(TypeCheckError):
    """``typing.Literal`` requested a dependency that doesn't exist in local state

    Args:
        target_cls (Type): the type which contains the problem ``typing.Literal``
        missing_dependency (Dict[str, str]): the dependency requests that didn't
            exist in the local state
    """

    def __init__(self, target_cls: Type, missing_dependency: Dict[str, str]) -> None:
        msg = [
            f"{target_cls.cls}'s' ``typing.Literal`` requested",
            "dependencies that were not in the local state.",
            "The following dependency requests were not in the local state:",
            *[f"\t- {k}: {v}" for k, v in missing_dependency.items()],
        ]
        self.missing_dependency = missing_dependency
        super().__init__(target_cls, msg, child=None, obj=None)


class AnumWrongTypeError(WrongTypeError):
    """Error when provided an ``obj`` of the wrong type for a :class:`~adorn.unit.anum.Anum`

    Args:
        target_cls (Anum): the requested wrapped enumeration
        obj (Any): caller provided object, that is not of type
            ``target_cls``
    """  # noqa: B950

    def __init__(self, target_cls: "Anum", obj: Any) -> "TypeCheckError":
        super().__init__(str, obj)
        self.target_cls = target_cls
        self.msg[0] = f"For the Anum, {self.target_cls}, " + self.msg[0]


class AnumMemberError(TypeCheckError):
    """The str specified doesn't match any of the members of ``target_cls``

    Args:
        target_cls (Anum): the requested wrapped enumeration
        obj (str): requested member that does not exist
    """

    def __init__(self, target_cls: "Anum", obj: str) -> None:
        self.members = target_cls.__members__
        msg = [
            f"For {target_cls}, a str specified a member,",
            f"{obj}, which is not an acceptable option",
            f"for {target_cls}.  The valid members",
            f"of {target_cls} are:",
            *[f"\t- {k}" for k in self.members.keys()],
        ]
        super().__init__(target_cls, msg, child=None, obj=obj)


class UserDictError(TypeCheckError):
    """Report an error during the from_obj step of :class:`~adorn.alter.dict_alter.UserDictAlter`

    Args:
        alter_type (Type): the type of ``Alter`` being performed
        target_cls (Parameter): the parameter the alter was trying to convert
            the ``obj`` into
        obj (Any): the alter request that produced an exception
        exception (Exception): the ``Exception`` produced from trying to convert ``obj``
            into the requested data.

    """  # noqa: B950

    def __init__(
        self, alter_type: Type, target_cls: "Parameter", obj: Any, exception: Exception
    ) -> None:
        self.alter_type = alter_type
        self.exception = exception
        msg = [
            f"An Alter of type {alter_type} was requested for",
            f"a parameter named {target_cls.parameter_name} of type {target_cls.cls}",
            "but an exception was caused when converting the obj:",
            f"{obj}",
            f"to type {target_cls.cls}.  The exception was:",
            f"{self.exception}",
        ]
        super().__init__(target_cls, msg, child=None, obj=obj)

    def __eq__(self, other) -> bool:
        if type(self) != type(other):
            return False
        return all(
            [
                self.target_cls == other.target_cls,
                self.msg[:-1] == other.msg[:-1],
                self.child == other.child,
                self.obj == other.obj,
                isinstance(self.exception, type(other.exception)),
            ]
        )



# Copyright 2021 Jacob Baumbach
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A container of information relevant for constructing an object."""
import inspect
from dataclasses import dataclass
from typing import Callable
from typing import Dict
from typing import Type
from typing import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
    from adorn.unit.complex import _T, _UnitT


@dataclass
class Constructor:
    """Container holding the class and relevant metadata to construct an object.

    Args:
        subclass (_UnitT): The ``type`` to generate construction metadata for
    """

    subclass: "_UnitT"

    def __post_init__(self):
        self.constructor_type = getattr(self.subclass, "constructor_type", None)
        if hasattr(self.subclass, "by_name"):
            self.constructor = self.subclass.by_name()
            constructor = (
                self.constructor
                if self.constructor.constructor_name is None
                else getattr(self.constructor, self.constructor.constructor_name)
            )
        else:
            self.constructor = self.subclass
            constructor = self.constructor
        self.parameters = self.infer_params(self.subclass, constructor)
        self.parameter_order = getattr(self.subclass, "parameter_order", None)

    def infer_params(
        self, target_class: Type["_T"], constructor: Callable[..., "_T"]
    ) -> Dict[str, inspect.Parameter]:
        """Generate a dict containing all the parameters needed to instantiate ``subclass``.

        This method is recursive and will convert ``**kwargs``, specified in the constructor,
        to the relevant parameters of its parent class(es), given the parent(s)
        subclass :class:`~adorn.unit.complex.Complex`

        Args:
            target_class (Type["_T"]): The class you are interegating the constructor of.
                This argument is used to explode the ``**kwargs``
            constructor (Callable[..., "_T"]): The method, whose parameters will be added
                to a dictionary

        Returns:
            Dict[str, inspect.Parameter]: constructor parameters as a dict
                keys are the parameter names and values are the parameter metadata
        """  # noqa:B950
        # circular imports
        from adorn.unit.complex import Complex

        signature = inspect.signature(constructor)
        parameters = dict(signature.parameters)

        parameters.pop("self", None)

        has_kwargs = False
        var_positional_key = None
        for param in parameters.values():
            if param.kind == param.VAR_KEYWORD:
                has_kwargs = True
            elif param.kind == param.VAR_POSITIONAL:
                var_positional_key = param.name

        if var_positional_key:
            del parameters[var_positional_key]

        if not has_kwargs:
            return parameters
        # probably only want to do this when the associated type is inspect._empty
        parameters.pop("kwargs")
        # "mro" is "method resolution order".  The first one is the current class,
        # the next is the first superclass, and so on.  We take the first
        # superclass we find that inherits from Complex.
        super_class = None
        # we could probably use the Complex's parent method
        # to help filter the mro output
        for super_class_candidate in target_class.mro()[1:]:
            if issubclass(super_class_candidate, Complex):
                super_class = super_class_candidate
                break
        if super_class:
            super_parameters = self.infer_params(super_class, super_class.by_name())
        else:
            super_parameters = {}

        return {
            **super_parameters,
            **parameters,
        }  # Subclass parameters overwrite superclass ones



# Copyright 2021 Jacob Baumbach
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Representation of ``enum.Enum``"""
from enum import Enum
from inspect import isclass
from typing import Any
from typing import Callable
from typing import ClassVar
from typing import DefaultDict
from typing import Dict
from typing import Optional
from typing import Type
from typing import TYPE_CHECKING
from typing import TypeVar

if TYPE_CHECKING:  # pragma: no cover
    from adorn.orchestrator.orchestrator import Orchestrator

from adorn.exception.type_check_error import (
    AnumMemberError,
    AnumWrongTypeError,
    TypeCheckError,
    UnRepresentedTypeError,
)
from adorn.unit.unit import Unit


_E = TypeVar("_E", bound=Enum)


class Anum(Unit):
    """A wrapper around :class:`~enum.Enum`

    The members of the enumeration are added to
    :attr:`~adorn.unit.anum.Anum._registry` and are accessible
    by specifying the enumeration they belong to and the ``str``
    representation of their name

    .. code-block:: python

        from collections import defaultdict
        from enum import auto, Enum

        from adorn.orchestrator.base import Base
        from adorn.unit.anum import Anum

        class Example(Anum):
            _registry = defaultdict(dict)

        @Example.register()
        class AB(Enum):
            a = auto
            b = auto

        base = Base([Example()])
        assert base.from_obj(AB, "a") == AB.a

    """  # noqa: RST304

    _registry: ClassVar[DefaultDict[type, Dict[str, type]]]

    @classmethod
    def register(cls) -> Callable[[Type[_E]], Type[_E]]:  # pragma: no cover
        """Add :class:`~enum.Enum` and it's members to the :attr:`~adorn.unit.anum.Anum._registry`

        Returns:
            Callable[[Type[_E]], Type[_E]]: funcion that adds the decorated
                :class:`~enum.Enum` and its members to the ``_registry``
        """  # noqa: B950, D202, RST304

        def add_members(subclass: Type[_E]) -> Type[_E]:
            registry = cls._registry[subclass]
            for k, v in subclass.__members__.items():
                registry[k] = v
            return subclass

        return add_members

    def get(
        self, key: Any, orchestrator: "Orchestrator"
    ) -> Optional[type]:  # pragma: no cover
        """Return the relevant ``Anum`` for the `key`

        Args:
            key (Any): the Type that may be a subclass of the ``Unit``
            orchestrator (Orchestrator): container of all types, not used

        Returns:
            Optional[type]: if ``key`` is an ``Anum`` then return it,
                otherwise ``None``
        """
        if self.contains(key, orchestrator):
            return key

    def contains(
        self, obj: Any, orchestrator: "Orchestrator"
    ) -> bool:  # pragma: no cover
        """Check if `obj` is registered to ``Anum``

        Args:
            obj (Any): the Type that may be a subclass of ``Anum``
            orchestrator (Orchestrator): container of all types, not used

        Returns:
            bool: if ``True`` then ``obj`` is represented by ``Anum``,
                otherwise ``False``
        """
        return isclass(obj) and issubclass(obj, Enum) and obj in self._registry

    def type_check(
        self, target_cls: Type, orchestrator: "Orchestrator", obj: Any
    ) -> Optional[TypeCheckError]:  # pragma: no cover
        """Ensure `obj` can be converted to the type ``target_cls``

        Exceptions:
            - :class:`~adorn.exception.type_check_error.UnRepresentedTypeError`:
                ``obj`` did not map to any of the types contained in
                the ``_registry``
            - :class:`~adorn.exception.type_check_error.AnumWrongTypeError`:
                ``obj`` was not of type ``str``
            - :class:`~adorn.exception.type_check_error.AnumMemberError`:
                ``obj`` specified a value that did not map to one of the
                members of ``target_cls``

        Args:
            target_cls (Type): the target Type, the given ``obj`` will be checked
                against
            orchestrator (Orchestrator): container of all types, not used
            obj (Any): an instance, that will be checked to see if it
                can be converted to an instance of ``target_cls``

        Returns:
            Optional[TypeCheckError]: if ``TypeCheckError`` then there was
                an issue which would prevent converting ``obj`` to an instance
                of ``target_cls``, otherwise ``None``
        """
        if not self.contains(target_cls, orchestrator):
            return UnRepresentedTypeError(target_cls=target_cls, unit=self, obj=obj)

        if not isinstance(obj, str):
            return AnumWrongTypeError(target_cls=target_cls, obj=obj)

        if obj not in self._registry[target_cls]:
            return AnumMemberError(target_cls=target_cls, obj=obj)

    def from_obj(self, target_cls: Type, orchestrator: "Orchestrator", obj: Any) -> _E:
        """Convert `obj` to an instance of the type ``target_cls``

        Args:
            target_cls (Type): the target Type, the given ``obj`` will be converted
                into
            orchestrator (Orchestrator): container of all types, not used
            obj (Any): an instance, that will be converted to an
                instance of ``target_cls``

        Returns:
            _E: a member of ``target_cls``
        """
        return self._registry[target_cls][obj]



# Copyright 2021 Jacob Baumbach
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Generic exception for a misconfigured object."""


class ConfigurationError(Exception):
    """The exception raised by any object when it's misconfigured.

    (e.g. missing properties, invalid properties, unknown properties).

    .. note::

        This ``Exception`` should typically be avoided, and instead an
        exception that subclasses
        :class:`~adorn.exception.type_check_error.TypeCheckError`
        or a new custom ``Exception`` should be used.  These alternate
        ``Exception`` objects contain more information, and are therefore
        more useful for the caller.
    """

    def __init__(self, message: str):
        super().__init__()
        self.message = message

    def __str__(self):
        return self.message



# Copyright 2021 Jacob Baumbach
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Base Class for altering a config during type checking or object creation"""
from typing import Any
from typing import Optional
from typing import Type
from typing import TYPE_CHECKING
from typing import Union

from adorn.exception.type_check_error import TypeCheckError

if TYPE_CHECKING:  # pragma: no cover
    from adorn.orchestrator.orchestrator import Orchestrator


class Alter:
    """Base class that enables a config to be altered during type checking or object creation"""  # noqa: B950

    def a_contains(
        self,
        target_cls: Type,
        orchestrator: "Orchestrator",
        obj: Any,
    ) -> bool:  # pragma: no cover
        """Check if the application of an ``Alter`` is required

        Args:
            target_cls (Type): the target Type, the given ``obj`` will be converted
                into or checked against
            orchestrator (Orchestrator): container of all types, typically used
                to recurse down nested types
            obj (Any): an instance, that will be checked to see if it can be
                converted into the ``target_cls`` or will be converted to an
                instance of ``target_cls``

        Raises:
            NotImplementedError: This method must be overriden by a subclass
        """
        raise NotImplementedError

    def a_get(
        self,
        target_cls: Type,
        orchestrator: "Orchestrator",
        obj: Any,
    ) -> Optional["Alter"]:  # pragma: no cover
        """Capture the ``Alter`` associated with the arguments

        Args:
            target_cls (Type): the target Type, the given ``obj`` will be converted
                into or checked against
            orchestrator (Orchestrator): container of all types, typically used
                to recurse down nested types
            obj (Any): an instance, that will be checked to see if it can be
                converted into the ``target_cls`` or will be converted to an
                instance of ``target_cls``

        Raises:
            NotImplementedError: This method must be overriden by a subclass
        """
        raise NotImplementedError

    def alter_obj(
        self,
        target_cls: Type,
        orchestrator: "Orchestrator",
        obj: Any,
    ) -> Union[Any, TypeCheckError]:  # pragma: no cover
        """Apply an alteration to an object

        Args:
            target_cls (Type): the target Type, the given ``obj`` will be converted
                into or checked against
            orchestrator (Orchestrator): container of all types, typically used
                to recurse down nested types
            obj (Any): an instance, that will be checked to see if it can be
                converted into the ``target_cls`` or will be converted to an
                instance of ``target_cls``

        Raises:
            NotImplementedError: This method must be overriden by a subclass
        """
        raise NotImplementedError

    def log_instance(
        self,
        target_cls: Type,
        orchestrator: "Orchestrator",
        obj: Any,
        instance: Any,
    ) -> None:  # pragma: no cover
        """Apply logging to an instantiate object

        Args:
            target_cls (Type): the target Type, the given ``obj`` was converted
                into
            orchestrator (Orchestrator): container of all types, typically used
                to recurse down nested types
            obj (Any): an instance, that was converted to an
                instance of ``target_cls``
            instance (Any): the output of converting ``obj`` into an instance
                of ``target_cls``

        Raises:
            NotImplementedError: This method must be overriden by a subclass
        """
        raise NotImplementedError



# Copyright 2021 Jacob Baumbach
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Vanilla coordinator"""
from collections import OrderedDict
from typing import Any
from typing import List
from typing import Optional
from typing import Type
from typing import Union

from adorn.alter.alter import Alter
from adorn.exception.type_check_error import TypeCheckError
from adorn.orchestrator.orchestrator import Orchestrator
from adorn.unit.unit import Unit


class Base(Orchestrator):
    """Vanilla coordinator between different :class:`~adorn.unit.unit.Unit`

    ``Base`` finds the first :class:`~adorn.unit.unit.Unit` specified in
    the list of :class:`~adorn.unit.unit.Unit` that contains a type
    and delegates the action to the given :class:`~adorn.unit.unit.Unit`.
    Therefore, the earlier a :class:`~adorn.unit.unit.Unit` appears in
    ``units`` the higher priority it has in the orchestrator.

    Args:
        units (List[Unit]): a collection of collection of types
        alters (List[Alter]): a collection of ways to dynamically alter
            a configuration

    Attributes:
        unit_dict (OrderedDict[Type, Unit]): a mapping from a type of a
            :class:`~adorn.unit.unit.Unit` to an instance of the
            :class:`~adorn.unit.unit.Unit`
    """

    def __init__(self, units: List[Unit], alters: Optional[List[Alter]] = None) -> None:
        super().__init__()
        self.units = units
        self.alters = alters or []
        self.unit_dict = OrderedDict([(type(i), i) for i in self.units])

    def contains(self, cls: Type) -> bool:
        """Check if ``Type`` is contained in one of the :class:`~adorn.unit.unit.Unit` objects

        Args:
            cls (Type): a type that is being checked to see if
                it exists in one of the :class:`~adorn.unit.unit.Unit`

        Returns:
            bool: if ``True`` a :class:`~adorn.unit.unit.Unit`
                contained in the orchestrator contains
                the provided type
        """  # noqa: B950
        return any(i.contains(cls, self) for i in self.units)

    def get(self, cls: Type) -> Unit:
        """Finds the :class:`~adorn.unit.unit.Unit` associated with `cls`

        .. note::

            The first :class:`~adorn.unit.unit.Unit` will be returned,
            where order is determined by a
            :class:`~adorn.unit.unit.Unit` location in ``units``.

        Args:
            cls (Type): a type that you want the associated ``Unit`` of

        Returns:
            Unit: The collection of types that contains ``cls``

        Raises:
            TypeCheckError: ``cls`` cannot be represented by any of
                the :class:`~adorn.unit.unit.Unit` contained in the
                orchestrator
        """
        for i in self.units:
            if i.contains(cls, self):
                return i
        ru_str = "\n".join([f"{type(i)}" for i in self.units])
        raise TypeCheckError(
            target_cls=cls,
            msg=[
                f"{cls} is not supported by any of the following Unit's:",
                f"{ru_str}",
            ],
        )

    def a_get(self, cls: Type, obj: Any) -> List[Alter]:
        """Finds all the :class:`~adorn.alter.alter.Alter` associated with `cls`

        .. note::

            The list of :class:`~adorn.alter.alter.Alter` will be returned,
            in the order :class:`~adorn.alter.alter.Alter` were passed.

        Args:
            cls (Type): a type that you want to check or convert the ``obj`` into
            obj (Any): the object to be type checked or converted into ``cls``

        Returns:
            List[Alter]: the classes requested to perform a dynamic alteration
                to the ``obj``
        """
        return [i for i in self.alters if i.a_contains(cls, self, obj)]

    def alter_obj(self, cls: Type, obj: Any) -> Union[Any, TypeCheckError]:
        """Potentially perform a dynamic alteration to an object

        Args:
            cls (Type): a type that you want to check or convert the ``obj`` into
            obj (Any): the object to be type checked or converted into ``cls``

        Returns:
            Union[Any, TypeCheckError]: Either an unaltered ``obj`` is returned, when no
                :class`~adorn.alter.alter.Alter` in ``alters`` contains the
                given ``cls`` and ``obj``, an altered ``obj`` is returned when the
                given ``cls`` and ``obj`` is contained by an
                :class`~adorn.alter.alter.Alter` in ``alters``, or a ``TypeCheckError``,
                when there was an issue performing a dynamic alteration to the ``obj``
        """
        alter_list = self.a_get(cls, obj)
        for alter in alter_list:
            obj = alter.alter_obj(cls, self, obj)
        return obj

    def log_instance(self, cls: Type, obj: Any, instance: Any) -> None:
        """Potentially perform a dynamic logging of instantiated instances

        Args:
            cls (Type): a type that you converted the ``obj`` into
            obj (Any): the object converted into ``cls``
            instance (Any): the object created from converting ``obj`` into an
                instance of ``cls``
        """
        alter_list = self.a_get(cls, obj)
        for alter in alter_list:
            obj = alter.log_instance(cls, self, obj, instance)

    def type_check(self, cls: Type, obj: Any) -> Optional[TypeCheckError]:
        """Check if `obj` can be converted to type of `cls`

        First, the code gets the :class:`~adorn.unit.unit.Unit` that
        contains ``cls``.  Then the type checking is delegated to the
        relevant :class:`~adorn.unit.unit.Unit`.

        Args:
            cls (Type): the type ``obj`` would be converted into
            obj (Any): an instance to be converted into type ``cls``

        Returns:
            Optional[TypeCheckError]: if ``None``, ``obj`` can be converted to an
                instance of ``cls``, otherwise an error is returned explaining
                why an ``obj`` cannot be converted to an instance of ``cls``
        """
        obj = self.alter_obj(cls, obj)
        if isinstance(obj, TypeCheckError):
            return obj
        ru = self.get(cls)
        return ru.type_check(cls, self, obj)

    def from_obj(self, cls: Type, obj: Any) -> Any:
        """Generate an instance of type `cls` from `obj`

        First, the code gets the :class:`~adorn.unit.unit.Unit` that
        contains ``cls``.  Then the instantiating is delegated to the
        relevant :class:`~adorn.unit.unit.Unit`.

        Args:
            cls (Type): the type ``obj`` would be converted into
            obj (Any): an instance to be converted into type ``cls``

        Returns:
            Any: an instance of type ``cls``, that was created from ``obj``

        Raises:
            obj: obj is a ``TypeCheckError`` indicating there was an issue
                applying a dynamic alteration to ``obj``
        """
        obj = self.alter_obj(cls, obj)
        if isinstance(obj, TypeCheckError):
            raise obj
        ru = self.get(cls)
        output = ru.from_obj(cls, self, obj)
        self.log_instance(cls, obj, output)
        return output



# Copyright 2021 Jacob Baumbach
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Params: a glorified dictionary"""
import copy
import json
import logging
import zlib
from collections import OrderedDict
from collections.abc import MutableMapping
from os import PathLike
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import tomli
import yaml

from adorn.exception.configuration_error import ConfigurationError

logger = logging.getLogger(__name__)


class ParamsEncoder(json.JSONEncoder):
    """Custom Encoder for nested ``Params``"""

    def default(self, obj):
        """Ensure nested ``Params`` are serializable

        Args:
            obj: something trying to be serialized to disk

        Returns:
            Encoded version of the input
        """
        if isinstance(obj, Params):
            return obj.as_ordered_dict()
        return json.JSONEncoder.default(self, obj)


def infer_and_cast(value: Any) -> Any:  # noqa:C901
    """In some cases we'll be feeding params dicts to functions we don't own

    For example, PyTorch optimizers. In that case we can't use `pop_int`
    or similar to force casts (which means you can't specify `int` parameters
    using environment variables). This function takes something that looks JSON-like
    and recursively casts things that look like (bool, int, float) to
    (bool, int, float).
    """
    if isinstance(value, (int, float, bool)):
        # Already one of our desired types, so leave as is.
        return value
    elif isinstance(value, list):
        # Recursively call on each list element.
        return [infer_and_cast(item) for item in value]
    elif isinstance(value, dict):
        # Recursively call on each dict value.
        return {key: infer_and_cast(item) for key, item in value.items()}
    elif isinstance(value, str):
        # If it looks like a bool, make it a bool.
        if value.lower() == "true":
            return True
        elif value.lower() == "false":
            return False
        else:
            # See if it could be an int.
            try:
                return int(value)
            except ValueError:
                pass
            # See if it could be a float.
            try:
                return float(value)
            except ValueError:
                # Just return it as a string.
                return value
    else:
        raise ValueError(f"cannot infer type of {value}")


def unflatten(flat_dict: Dict[str, Any]) -> Dict[str, Any]:
    """Given a "flattened" dict with compound keys, e.g.

    ``{"a.b": 0}`` after unflatten, ``{"a": {"b": 0}}``
    """
    unflat: Dict[str, Any] = {}

    for compound_key, value in flat_dict.items():
        curr_dict = unflat
        parts = compound_key.split(".")
        for key in parts[:-1]:
            curr_value = curr_dict.get(key)
            if key not in curr_dict:
                curr_dict[key] = {}
                curr_dict = curr_dict[key]
            elif isinstance(curr_value, dict):
                curr_dict = curr_value
            else:
                raise ConfigurationError("flattened dictionary is invalid")
        if not isinstance(curr_dict, dict) or parts[-1] in curr_dict:
            raise ConfigurationError("flattened dictionary is invalid")
        curr_dict[parts[-1]] = value

    return unflat


def _is_dict_free(obj: Any) -> bool:
    """Returns False if obj is a dict, otherwise list with an element that _has_dict."""
    if isinstance(obj, dict):
        return False
    elif isinstance(obj, list):
        return all(_is_dict_free(item) for item in obj)
    else:
        return True


class Params(MutableMapping):
    """Represents a parameter dictionary with a history.

    Contains other functionality around parameter passing and validation for Adorn.
    There are currently two benefits of a `Params` object over a plain dictionary for
    parameter passing:

    1. We handle a few kinds of parameter validation, including making sure that
       parameters representing discrete choices actually have acceptable values,
       and making sure no extra parameters are passed.
    2. We log all parameter reads, including default values.  This gives a more
       complete specification of the actual parameters used than is given in a JSON
       file, because those may not specify what default values were used, whereas
       this will log them.
    """

    # This allows us to check for the presence of "None" as a default argument,
    # which we require because we make a distinction between passing a value of "None"
    # and passing no value to the default parameter of "pop".
    DEFAULT = object()

    def __init__(self, params: Dict[str, Any], history: str = "") -> None:
        self.params = _replace_none(params)
        self.history = history

    def pop(self, key: str, default: Any = DEFAULT, keep_as_dict: bool = False) -> Any:
        """Performs the functionality associated with dict.pop(key).

        Along with checking for returned dictionaries, replacing them
        with Param objects with an updated history (unless keep_as_dict is True, in
        which case we leave them as dictionaries). If `key` is not present in the
        dictionary, and no default was specified, we raise a `ConfigurationError`,
        instead of the typical `KeyError`.
        """  # noqa:D402
        if default is self.DEFAULT:
            try:
                value = self.params.pop(key)
            except KeyError:
                msg = f'key "{key}" is required'
                if self.history:
                    msg += f' at location "{self.history}"'
                raise ConfigurationError(msg) from KeyError
        else:
            value = self.params.pop(key, default)

        if keep_as_dict or _is_dict_free(value):
            logger.info(f"{self.history}{key} = {value}")
            return value
        else:
            return self._check_is_dict(key, value)

    def pop_int(self, key: str, default: Any = DEFAULT) -> Optional[int]:
        """Performs a pop and coerces to an int."""
        value = self.pop(key, default)
        if value is None:
            return None
        else:
            return int(value)

    def pop_float(self, key: str, default: Any = DEFAULT) -> Optional[float]:
        """Performs a pop and coerces to a float."""
        value = self.pop(key, default)
        if value is None:
            return None
        else:
            return float(value)

    def pop_bool(self, key: str, default: Any = DEFAULT) -> Optional[bool]:
        """Performs a pop and coerces to a bool."""
        value = self.pop(key, default)
        if value is None:
            return None
        elif isinstance(value, bool):
            return value
        elif value == "true":
            return True
        elif value == "false":
            return False
        else:
            raise ValueError(f"Cannot convert variable to bool: {value}")

    def get(self, key: str, default: Any = DEFAULT) -> Any:
        """Performs the functionality associated with dict.get(key).

        Also checks for returned dicts and returns a Params object in
        their place with an updated history.
        """  # noqa:D402
        default = None if default is self.DEFAULT else default
        value = self.params.get(key, default)
        return self._check_is_dict(key, value)

    def pop_choice(
        self,
        key: str,
        choices: List[Any],
        default_to_first_choice: bool = False,
        allow_class_names: bool = True,
    ) -> Any:
        """Gets the value of `key` in the `params` dictionary,

        Ensuring that the value is one of the given choices. Note that
        this `pops` the key from params, modifying the dictionary,
        consistent with how parameters are processed in this codebase.

        Args:
            key (str): Key to get the value from in the param dictionary
            choices (List[Any]): A list of valid options for values corresponding to `key`.
                For example, if you're specifying the type of encoder to use for some part
                of your model, the choices might be the list of encoder classes we know
                about and can instantiate.  If the value we find in the param dictionary
                is not in `choices`, we raise a `ConfigurationError`, because
                the user specified an invalid value in their parameter file.
            default_to_first_choice (bool): If this is `True`, we allow the `key` to not be
                present in the parameter dictionary.  If the key is not present, we will
                use the return as the value the first choice in the `choices` list.
                If this is `False`, we raise a `ConfigurationError`, because
                specifying the `key` is required (e.g., you `have` to specify
                your model class when running an experiment, but you can
                feel free to usedefault settings for encoders if you want).
            allow_class_names (bool): If this is `True`, then we allow unknown
                choices that look like fully-qualified class names.
                This is to allow e.g. specifying a model type
                as my_library.my_model.MyModel and importing it on the fly. Our check for
                "looks like" is extremely lenient and consists of checking that the value
                contains a '.'.

        Returns:
            Any: value which was one of the listed ``choices``

        Raises:
            ConfigurationError: requested a key that DNE
        """  # noqa: B950
        default = choices[0] if default_to_first_choice else self.DEFAULT
        value = self.pop(key, default)
        ok_because_class_name = allow_class_names and "." in value
        if value not in choices and not ok_because_class_name:
            key_str = self.history + key
            message = (
                f"{value} not in acceptable choices for {key_str}: {choices}. "
                "You should either use the --include-package flag to make "
                "sure the correct module is loaded, or use a fully qualified "
                "class name in your config file like "
                """{"model": "my_module.models.MyModel"} """
                "to have it imported automatically."
            )
            raise ConfigurationError(message)
        return value

    def as_dict(
        self, quiet: bool = False, infer_type_and_cast: bool = False
    ) -> Dict[str, Any]:
        """Sometimes we need to just represent the parameters as a dict.

        For instance when we pass them to PyTorch code.

        Args:
            quiet (bool): Whether to log the parameters before
                returning them as a dict.
            infer_type_and_cast (bool): If True, we infer types and cast
                (e.g. things that look like floats to floats).

        Returns:
            Dict[str, Any]: dict containing the content of ``Params``
        """
        if infer_type_and_cast:
            params_as_dict = infer_and_cast(self.params)
        else:
            params_as_dict = self.params

        if quiet:
            return params_as_dict

        def log_recursively(parameters, history):
            for key, value in parameters.items():
                if isinstance(value, (dict, Params)):
                    new_local_history = history + key + "."
                    log_recursively(value, new_local_history)
                else:
                    logger.info(f"{history}{key} = {value}")

        log_recursively(self.params, self.history)
        return params_as_dict

    def as_flat_dict(self) -> Dict[str, Any]:
        """Returns the parameters of a flat dictionary from keys to values.

        Nested structure is collapsed with periods.
        """
        flat_params = {}

        def recurse(parameters, path):
            for key, value in parameters.items():
                newpath = path + [key]
                if isinstance(value, (dict, Params)):
                    recurse(value, newpath)
                else:
                    flat_params[".".join(newpath)] = value

        recurse(self.params, [])
        return flat_params

    def duplicate(self) -> "Params":
        """Uses `copy.deepcopy()` to create a duplicate (but fully distinct)

        copy of these Params.
        """  # noqa:D402
        return copy.deepcopy(self)

    def assert_empty(self, class_name: str) -> None:
        """Raises a `ConfigurationError` if `self.params` is not empty.

        We take `class_name` as an argument so that the error message
        gives some idea of where an error happened, if there was one.
        `class_name` should be the name of the `calling` class,
        the one that got extra parameters (if there are any).
        """
        if self.params:
            raise ConfigurationError(
                "Extra parameters passed to {}: {}".format(class_name, self.params)
            )

    def __getitem__(self, key) -> Any:
        if key in self.params:
            return self._check_is_dict(key, self.params[key])
        else:
            raise KeyError

    def __setitem__(self, key, value) -> None:
        self.params[key] = value

    def __delitem__(self, key) -> None:
        del self.params[key]

    def __iter__(self):
        return iter(self.params)

    def __len__(self) -> int:
        return len(self.params)

    def _check_is_dict(self, new_history, value) -> Union["Params", List["Params"]]:
        if isinstance(value, dict):
            new_history = self.history + new_history + "."
            return Params(value, history=new_history)
        if isinstance(value, list):
            value = [
                self._check_is_dict(f"{new_history}.{i}", v)
                for i, v in enumerate(value)
            ]
        return value

    @classmethod
    def from_file(
        cls,
        params_file: Union[str, PathLike],
    ) -> "Params":
        """Load a `Params` object from a configuration file.

        Args:
            params_file (str): The path to the configuration file to load.

        Returns:
            Params: content of the file wrapped in a ``Params``
        """  # noqa: DAR401
        params_file = str(params_file)
        file_dict = None
        if params_file.endswith("json"):
            with open(params_file) as file_handle:
                file_dict = json.loads(file_handle.read())
        elif params_file.endswith("yaml"):
            with open(params_file) as file_handle:
                file_dict = yaml.safe_load(file_handle)
        elif params_file.endswith("toml"):
            with open(params_file, "rb") as file_handle:
                file_dict = tomli.load(file_handle)

        if file_dict is None:
            raise Exception("Only supported file formats are json, yaml and toml")
        return cls(file_dict)

    def to_file(self, params_file: str) -> None:
        """Writes params object to disk as a json"""
        with open(params_file, "w") as handle:
            json.dump(self.as_ordered_dict(), handle, indent=4, cls=ParamsEncoder)

    def as_ordered_dict(self) -> OrderedDict:
        """Returns Ordered Dict of Params from list of partial order preferences."""
        params_dict = self.as_dict(quiet=True)

        def order_dict(dictionary):
            # Recursively orders dictionary according to scoring order_func
            result = OrderedDict()
            for key, val in sorted(dictionary.items(), key=lambda item: item[0]):
                result[key] = (
                    order_dict(val) if isinstance(val, (dict, Params)) else val
                )
            return result

        return order_dict(params_dict)

    def get_hash(self) -> str:
        """Returns a hash code representing the current state of this `Params` object.

        We don't want to implement `__hash__` because that has deeper python
        implications (and this is a mutable object), but this will give you a
        representation of the current state. We use `zlib.adler32` instead of Python's
        builtin `hash` because the random seed for the latter is reset on each new
        program invocation, as discussed here:
        https://stackoverflow.com/questions/27954892/deterministic-hashing-in-python-3.
        """
        dumped = json.dumps(self.params, sort_keys=True)
        hashed = zlib.adler32(dumped.encode())
        return str(hashed)

    def __str__(self) -> str:
        return f"{self.history}Params({self.params})"

    def left_merge(self, rhs: Union["Params", Dict[str, Any]]) -> Dict[str, Any]:
        """Inject values from a new config into the current config

        Args:
            rhs (Union[Params, Dict[str, Any]]): configuration, where its values
                will be injected into the current config

        Returns:
            Dict[str, Any]: flattened version of the original config that includes
                values from the passed config, ``rhs``
        """
        self_flat_dict = self.as_flat_dict()
        rhs_flat_dict = (
            rhs.as_flat_dict()
            if isinstance(rhs, Params)
            else Params(rhs).as_flat_dict()
        )
        for k, v in rhs_flat_dict.items():
            self_flat_dict[k] = v
        return self_flat_dict


def _replace_none(params: Any) -> Any:
    if params == "None":
        return None
    elif isinstance(params, dict):
        for key, value in params.items():
            params[key] = _replace_none(value)
        return params
    elif isinstance(params, list):
        return [_replace_none(value) for value in params]
    return params
